#!/usr/bin/env python3
#
# By Julian Catchen <jcatchen@illinois.edu>, Copyright 2019-2021
#

from datetime import datetime
import math
import argparse
import shlex
import os
import sys
import gzip

#
# Global configuration variables.
#
scaf_agp_path = ""
chr_agp_path  = ""
ctg_agp_path  = ""
fasta_path    = ""
gff_path      = ""
gtf_path      = ""
changes_path  = ""
out_path      = ""
out_prefix    = ""
chr_whitelist = ""
line_len      = 60
scaf_gap_size = 500

unordered_scaffolds_concatenated = False
synthetic_chr = ""

defined_operations   = {"invert", "invert_scaffold", "insert_before", "insert_after", "split_after", "fuse", "rename"}
defined_ann_elements = {"gene", "mRNA", "CDS", "transcript", "five_prime_UTR", "three_prime_UTR", "five_prime_utr", "three_prime_utr",
                        "exon", "intron", "start_codon", "stop_codon",
                        "miRNA", "ncRNA_gene", "snRNA"}

class AGPComp():
    """ 
    Object to hold attributes of an AGP object (as read from an AGP file). 
    """


class AnnAtr():
    """
    Hold a single key/value pair from the attributes field of a GFF or GTF annotation file.
    """
    def __init__(self, key, value):
        self.key   = key
        self.value = value
    def __str__(self):
        s = self.key + ": '" + self.value + "'"
        return s

class AnnComp():
    """
    Object to hold gene annotations as parsed from a GFF or GTF file.
       https://github.com/The-Sequence-Ontology/Specifications/blob/master/gff3.md
    """
    def __init__(self, chr, src, type, start, end, score, strand, phase):
        self.chr     = chr
        self.src     = src
        self.type    = type
        self.sbp     = start
        self.ebp     = end
        self.score   = score
        self.strand  = strand
        self.phase   = phase
        self.attribs = []

    def parse_attributes(self, attrib_str, field_delim, keyval_delim):
        #
        # Split the key/value pairs, delimited by ';' while handling quotes (potentially with semicolons within the quotes).
        #
        attribs = []
        stop = len(attrib_str)
        p = 0
        q = p + 1
        while q < stop:
            while attrib_str[q] != field_delim and q < stop:
                if attrib_str[q] == "\"":
                    q += 1
                    while q < stop and attrib_str[q] != "\"":
                        q += 1
                elif attrib_str[q] == "\'":
                    q += 1
                    while q < stop and attrib_str[q] != "\'":
                        q += 1
                if q >= stop:
                    print("Error: unclosed quote in attributes string, '{}'".format(attrib_str), file=sys.stderr)
                    return -1
                q += 1
            # Split off segment
            att = attrib_str[p:q]
            att = att.strip(' ')
            if len(att) != 0:
                attribs.append(att)
            p = q + 1
            q = p + 1
        #
        # Split each key/value pair by keyval_delim (space in GTF, '=' in GFF).
        # If there is no delimeter, assume this is an unlabled gene_id as produced by Augustus/Braker.
        #
        for att in attribs:
            stop = len(att)
            p = 0
            q = p + 1
            while q < stop and att[q] != keyval_delim:
                if att[q] == "\"":
                    q += 1
                    while q < stop and att[q] != "\"":
                        q += 1
                elif att[q] == "\'":
                    q += 1
                    while q < stop and att[q] != "\'":
                        q += 1
                if q >= stop:
                    print("Error: unclosed quote in {} component, '{}' of attributes string, '{}'".format(att, attrib_str), file=sys.stderr)
                    return -1
                q += 1
            if q >= stop:
                key = "gene_id"
                val = att
            else:
                key = att[p:q].lower()
                val = att[q:]
                val = val.strip(" \"\'")
            self.attribs.append(AnnAtr(key, val))
        return 1


class CtgComp():
    """ 
    Object to hold attributes of a contig or inter-contig gap component (housed 
    by a scaffold or chromosome object). 
    """
    def __init__(self, chr, type, part, start, end):
        self.chr      = chr
        self.type     = type
        self.part     = part
        self.chr_sbp  = start # Chromosome starting basepair that contains this contig.
        self.chr_ebp  = end
        self.len      = self.chr_ebp - self.chr_sbp + 1
        self.name     = "gap"    # Name of contig if not a gap.
        self.sbp      = 1        # Starting basepair of this contig.
        self.ebp      = self.len # Ending basepair of this contig (defaults to full lenght of contig, but may only be a portion of the contig).
        self.scaf_sbp = 0        # Scaffold starting basepair that contains this contig.
        self.scaf_ebp = 0        # Scaffold ending basepair that contains this contig.
        self.scaf     = "gap"    # Name of scaffold that contains this contig, if it is not a gap.
    def __str__(self):
        t = "gap" if self.type ==   "N" else self.name
        f = ""    if self.scaf == "gap" else self.scaf + ", "
        s = (self.chr + ", " + '{:>12}'.format(t) +
             " [" + '{:>14}'.format(f) + "part: " + str(self.part) + "] " +
             str(self.chr_sbp) + " - " + str(self.chr_ebp) + " (" + str(self.len) + "bp)")
        return s

class ChrComp():
    """ 
    Object to hold attributes of a scaffold or inter-scaffold gap component (housed 
    by a chromosome object).
    """
    def __init__(self, chr, type, name, start, end):
        self.chr   = chr
        self.type  = type  # "scaffold" or "gap"
        self.name  = name
        self.start = start # Chromosomal basepair coordinates
        self.end   = end
        self.ori   = "+"
        self.len   = self.end - self.start + 1
        self.seq   = ""  # Sequence extracted from FASTA file.
        self.comps = []  # Component parts, aka CtgComp, may be scaffolds or contigs.
        self.genes = []  # Contains genes annotated in this scaffold.
    def __str__(self):
        s = (self.chr + ", " + '{:>10}'.format(self.name) +
             " " + str(self.start) + " - " + str(self.end) + " (" + str(self.len) + "bp, orientation: " + self.ori + "), " +
             "number of components: " + str(len(self.comps)))
        return s

class Chr():
    """ Object to hold scaffolds and related data for each chromosome. """
    def __init__(self, chr):
        self.chr   = chr
        self.comps = [] # Scaffold components
        self.key   = {}
        self.genes = [] # Contains genes that cross scaffold boundaries
    def __str__(self):
        s = "\n"
        for c in self.comps:
            s += "{:>12}\t{}\t{}\t({}bp)\n".format(c.name, c.start, c.end, c.len)
        s += "\n"
        return s

class Op():
    """ Object to store operations we will perform to alter the genome. """
    def __init__(self, chr, source, operation, dest = "", ori = "+"):
        self.operation = operation # invert, insert_before, insert_after, split_after, fuse, or rename
        self.chr       = chr
        self.source    = source # Source scaffold/chromosome
        self.dest      = dest   # Destination scaffold (optional depending on operation)
        self.ori       = ori    # Orientation of destination scaffold (optional)
    def __str__(self):
        if self.operation == "invert_scaffold":
            s = "operation '{}', applied to chromosome {}, on contig '{}'".format(self.operation, self.chr, self.source)
        elif self.operation == "rename":
            s = "operation '{}', applied to chromosome {}, now known as '{}'".format(self.operation, self.chr, self.source)
        elif self.operation == "invert":
            s = "operation '{}', applied to chromosome {}".format(self.operation, self.chr)
        elif self.operation == "fuse":
            s = "operation '{}', applied to chromosome {}, on scaffold '{}'".format(self.operation, self.chr, self.source)
        else:
            s = "operation '{}', applied to chromosome {}, on scaffold '{}'".format(self.operation, self.chr, self.dest)
        return s
        
def parse_command_line():
    global scaf_agp_path, chr_agp_path, ctg_agp_path, fasta_path, gff_path, gtf_path
    global changes_path, out_path, out_prefix
    global unplaced_scaffolds_concatenated
    global scaf_gap_size
    global line_len
    global chr_whitelist

    desc = "Alter the structure of a genome assembly, including the AGP, GFF/GTF, and FASTA files by following defined changes listed in the ALTER file."
    p = argparse.ArgumentParser(description=desc)

    #
    # Add options.
    #
    p.add_argument("--chromosomes", type=str, dest="chr_agp_path", required=True,
                   help="AGP file describing the scaffolds within each chromosome.")
    p.add_argument("--scaffolds", type=str, dest="scaf_agp_path",
                   help="AGP file describing unplaced scaffolds. If specified, these will be added to the main chromosome AGP objects for processing.")
    p.add_argument("-f","--fasta", type=str, dest="fasta_path",
                   help="Path to assembly FASTA genome sequence file that will be modifide to implement structural changes.")
    p.add_argument("--gff", type=str, dest="gff_path",
                   help="Path to GFF file describing gene annotations that will be modified to implement structural changes.")
    p.add_argument("--gtf", type=str, dest="gtf_path",
                   help="Path to GTF file describing gene annotations that will be modified to implement structural changes.")
    p.add_argument("-a","--alter", type=str, dest="alter_path", required=True,
                   help="Path to file containing alterations to make to the genome.")
    p.add_argument("-o","--out", type=str, dest="out_path", required=True,
                   help="Write altered files to this path.")
    p.add_argument("--prefix", type=str, dest="file_prefix",
                   help="Name output files to OUT_PATH naming them with this common prefix [default: original file name date stamped].")
    p.add_argument("--synthetic", type=str, dest="synthetic_chr",
                   help="Sometimes unordered scaffolds are concatenated into a synthetic chromosome. If so, give the name of that chromosome here (must match FASTA/AGP IDs).")
    p.add_argument("--linelen", type=int, dest="linelen",
                   help="Line length for printing FASTA sequences [default: 60bp].")
    p.add_argument("--gapsize", type=int, dest="gapsize",
                   help="When inserting a new gap between scaffolds make it this size [default: 500bp].")
    p.add_argument("-w","--whitelist", type=str, dest="wl",
                   help="Process a single chromosome specified here.")
    p.add_argument("--contigs", type=str, dest="ctg_agp_path",
                   help="AGP file describing the contigs within each chromosome [rare; use in addition to the --chromosomes option].")    
    
    #
    # Parse the command line
    #
    args = p.parse_args()

    if args.scaf_agp_path != None:
        scaf_agp_path = args.scaf_agp_path
    if args.chr_agp_path != None:
        chr_agp_path = args.chr_agp_path
    if args.ctg_agp_path != None:
        ctg_agp_path = args.chr_agp_path
    if args.fasta_path != None:
        fasta_path = args.fasta_path
    if args.alter_path != None:
        changes_path = args.alter_path
    if args.gff_path != None:
        gff_path = args.gff_path
    if args.gtf_path != None:
        gtf_path = args.gtf_path
    if args.out_path != None:
        out_path = args.out_path
    if args.file_prefix != None:
        out_prefix = args.file_prefix
    if args.synthetic_chr != None:
        unordered_scaffolds_concatenated = True
        synthetic_chr = args.synthetic_chr
    if args.gapsize != None:
        scaf_gap_size = int(args.gapsize)
    if args.linelen != None:
        line_len = int(args.linelen)
    if args.wl != None:
        chr_whitelist = args.wl
        
    if len(chr_agp_path) == 0 or os.path.exists(chr_agp_path) == False:
        print("Error: you must specify a valid path to an AGP file describing the assembled chromosomes.", file=sys.stderr)
        p.print_help()
        sys.exit()

    if len(scaf_agp_path) > 0 and os.path.exists(scaf_agp_path) == False:
        print("Error: you must specify a valid path to an AGP file describing the contigs in the assembled scaffolds.", file=sys.stderr)
        p.print_help()
        sys.exit()

    if len(ctg_agp_path) > 0 and os.path.exists(ctg_agp_path) == False:
        print("Error: you must specify a valid path to an AGP file describing the contigs in the assembled chromosomes.", file=sys.stderr)
        p.print_help()
        sys.exit()

    if len(gtf_path) > 0 and len(gff_path) > 0:
        print("Error: you may specify a GFF file or GTF file, but not both.", file=sys.stderr)
        p.print_help()
        sys.exit()

    if len(gtf_path) > 0 and os.path.exists(gtf_path) == False:
        print("Error: unable to find GTF file, '{}'.".format(gtf_path), file=sys.stderr)
        p.print_help()
        sys.exit()

    if len(gff_path) > 0 and os.path.exists(gff_path) == False:
        print("Error: unable to find GFF file, '{}'.".format(gff_path), file=sys.stderr)
        p.print_help()
        sys.exit()

    if len(out_path) > 0 and os.path.exists(chr_agp_path) == False:
        print("Error: you must specify a valid path to write altered files to.", file=sys.stderr)
        p.print_help()
        sys.exit()
        
    if len(out_path) > 0 and out_path[-1] != "/":
        out_path += "/"

def load_agp_file(agp_path, agp, agp_key):
    #
    # Load an AGP file defining gaps in the assembly.
    #  Format: object     object_beg object_end part_number component_type component_id component_beg component_end  orientation
    #  Gaps:   object     object_beg object_end part_number N              gap_length   gap_type      linkage        -empty-
    #          JH556661.1 1          70736 1 W AGAJ01000001.1 1        70736    +
    #          JH556661.1 70737      70782 2 N 46             fragment yes
    #
    if agp_path[-3:] == ".gz":
        fh = gzip.open(agp_path, 'rb')
    else:
        fh = open(agp_path)

    for line in fh:
        line = line.strip('\n ')

        if line[0] == "#":
            continue
        
        parts = line.split('\t')

        comp = AGPComp()
        comp.chr   = parts[0]
        comp.start = int(parts[1])
        comp.end   = int(parts[2])
        comp.part  = int(parts[3])
        comp.type  = parts[4]
        comp.len   = int(parts[5]) if comp.type == "N" else (int(parts[7]) - int(parts[6]) + 1)

        if comp.type != "N":
            comp.sub     = parts[5]
            comp.sub_sbp = int(parts[6])
            comp.sub_ebp = int(parts[7])
            comp.sub_ori = parts[8]
        else:
            comp.sub = "gap"

        if comp.chr not in agp:
            agp[comp.chr] = []
        agp[comp.chr].append(comp)

        #
        # Populate a reverse lookup so we know where to look for the subcomponent.
        #
        if comp.sub != "gap":
            agp_key[comp.sub] = (comp.chr, len(agp[comp.chr]) - 1)

    fh.close()
    print("Parsed '" + agp_path + "', loaded structure information for " + str(len(agp)) + " assembly objects.", file=sys.stderr)

def load_gff_file(gff_path, ann):
    global defined_ann_elements
    ignored = set()
    #
    # Load a GFF3 file defining annotated genes/RNAs in the assembly.
    #  Format: seqid    source    type    start    end    score    strand    phase    attributes (semi-colon separated)
    # Attributes have 'id' and 'parent' labels. Each component has its own ID, but may also have a parent, e.g. a transcript's parent is a gene.
    #
    if gff_path[-3:] == ".gz":
        fh = gzip.open(gff_path, 'rb')
    else:
        fh = open(gff_path)

    parent_key = {}
    lineno     = 0

    for line in fh:
        lineno += 1
        line = line.strip('\n ')

        if line[0] == "#":
            continue
        
        parts = line.split('\t')

        if parts[2] not in defined_ann_elements:
            ignored.add(parts[2])
            continue

        comp = AnnComp(parts[0], parts[1], parts[2], int(parts[3]), int(parts[4]), parts[5], parts[6], parts[7])
        res  = comp.parse_attributes(parts[8], ";", "=")
        if res < 1:
            print("GFF parsing failure of '{}' on line {}".format(gff_path, lineno))
            exit()

        parent = ""
        id     = ""
        for attrib in comp.attribs:
            if attrib.key == "id":
                id = value
            if attrib.key == "parent":
                parent = value

        if len(id) > 0 and len(parent) > 0:
            if id not in parent_key:
                parent_key[id] = parent
            id = parent
            parent = ""

        if len(parent) > 0:
            if parent not in parent_key:
                continue
            id = parent_key[parent]    

        if id not in ann:
            ann[id] = []
        ann[id].append(comp)
        
    fh.close()

    print("Parsed '" + gff_path + "', loaded information for " + str(len(ann)) + " annotation objects.", file=sys.stderr)
    if len(ignored) > 0:
        print("    Ignored elements of the following types: " + ", ".join(ignored), file=sys.stderr)

def load_gtf_file(gtf_path, ann):
    global defined_ann_elements
    ignored = set()
    #
    # Load a GTF file defining annotated genes/RNAs in the assembly.
    #  Format: seqid    source    type    start    end    score    strand    phase    attributes (semi-colon separated)
    # Attributes are specific, e.g. 'transcript_id' or 'gene_id'. Gene components may not have a label (as produced by Augustus)
    # and may just have the plain gene name without a label.
    #
    if gtf_path[-3:] == ".gz":
        fh = gzip.open(gtf_path, 'rb')
    else:
        fh = open(gtf_path)

    parent_key = {}
    lineno     = 0

    for line in fh:
        lineno += 1
        line = line.strip('\n ')

        if line[0] == "#":
            continue
        
        parts = line.split('\t')

        if parts[2] not in defined_ann_elements:
            ignored.add(parts[2])
            continue

        comp = AnnComp(parts[0], parts[1], parts[2], int(parts[3]), int(parts[4]), parts[5], parts[6], parts[7])
        res  = comp.parse_attributes(parts[8], ";", " ")
        if res < 1:
            print("GTF parsing failure for '{}' on line {}".format(gtf_path, lineno))
            exit()

        parent = ""
        id     = ""

        for attrib in comp.attribs:
            if comp.type == "gene":
                if attrib.key == "gene_id":
                    id = attrib.value
            else:
                if attrib.key == "transcript_id":
                    id = attrib.value
                elif attrib.key == "gene_id":
                    parent = attrib.value

        if len(id) > 0 and len(parent) > 0:
            if id not in parent_key:
                parent_key[id] = parent
            id = parent
            parent = ""

        if len(parent) > 0:
            if parent not in parent_key:
                continue
            id = parent_key[parent]    

        if id not in ann:
            ann[id] = []
        ann[id].append(comp)
        
    fh.close()

    print("Parsed '" + gtf_path + "', loaded information for " + str(len(ann)) + " annotation objects.", file=sys.stderr)
    if len(ignored) > 0:
        print("    Ignored elements of the following types: " + ", ".join(ignored), file=sys.stderr)

def load_fasta_file(fasta_path, fasta):
    if fasta_path[-3:] == ".gz":
        gzipped = True;
        fh     = gzip.open(fasta_path, 'rb')
    else:
        gzipped = False;
        fh     = open(fasta_path)
    #
    # Load a FASTA file into a dictionary.
    #
    eof  = False
    line = fh.readline()
    if len(line) == 0:
        return
    line = line.strip('\n ')
    
    while len(line) == 0 or line[0] == "#":
        line = fh.readline()
        if len(line) == 0:
            return
        line = line.strip('\n ')

    while eof == False:
        id  = ""
        seq = ""

        if line[0] == ">":
            id   = line[1:]
            # Remove any comments from the FASTA ID line.
            pos  = id.find(' ')
            if pos >= 0:
                id = id[0:pos]
            line = ""

        while len(line) == 0 or line[0] != ">":
            line = fh.readline()
            if len(line) == 0:
                eof = True
                break
            line = line.strip('\n ')
            if len(line) == 0 or line[0] == "#":
                continue
            if line[0] != ">":
                seq += line
        #
        # Record the final sequence.
        #
        fasta[id] = seq

    fh.close()
    print("Parsed '" + fasta_path + "', loaded " + str(len(fasta)) + " sequences.", file=sys.stderr)
    
def merge_contig_scaffold_agp(chromosomes, chromosomes_key, chr_agp, scaf_agp, scaf_agp_key):
    #
    # Here we assume two AGP files have been specified:
    #   1. The first describes the scaffolds within each chromosome
    #   2. The second describes the contigs within each chromosome
    # This function is designed to merge these into one data structure containing chromosomes->scaffolds->contigs.
    #
    
    for chr in chr_agp:
        components    = []
        last_scaffold = ""
        scaf_bp       = 1

        for k in range(0, len(chr_agp[chr])):
            #
            # Create chromosome object, populate chromosome/contig information.
            #
            cc = CtgComp(chr, chr_agp[chr][k].type, chr_agp[chr][k].part, chr_agp[chr][k].start, chr_agp[chr][k].end)

            if chr_agp[chr][k].type != "N":
                cc.name = chr_agp[chr][k].sub
                cc.sbp  = chr_agp[chr][k].sub_sbp
                cc.ebp  = chr_agp[chr][k].sub_ebp
                #
                # Lookup scaffold information, if this is a new scaffold, restart the scaffold bp counter.
                #
                (scaffold, index) = scaf_agp_key[cc.name]
                if last_scaffold != scaffold:
                    scaf_bp       = 1
                    last_scaffold = scaffold

                cc.scaf     = scaffold
                cc.scaf_sbp = scaf_bp
                cc.scaf_ebp = scaf_bp + scaf_agp[scaffold][index].len - 1

            components.append(cc)
            scaf_bp += cc.len

        c             = Chr(chr)
        i             = 0
        max_index     = len(components) - 1
        last_scaffold = components[0].scaf

        #
        # Collapse scaffold-level components.
        #
        while i <= max_index:
            j = i
            while j < max_index and (components[j].type == "N" or components[j].scaf == last_scaffold):
                j += 1
            if j < max_index or components[j].scaf != last_scaffold:
                j -= 2
            chr_comp     = ChrComp(chr, "scaffold", components[i].scaf, components[i].chr_sbp, components[j].chr_ebp)
            chr_comp.ori = scaf_agp[scaffold][index].sub_ori

            # Collect all the contig/gap components and store them in this scaffold object.
            for k in range(i, j + 1):
                chr_comp.comps.append(components[k])
            #
            # Store this scaffold object and where to find it.
            #
            c.comps.append(chr_comp)
            c.key[chr_comp.name] = len(chr_comp.comps) - 1
            if chr_comp.name not in chromosomes_key:
                chromosomes_key[chr_comp.name] = []
            chromosomes_key[chr_comp.name].append( (chr, (len(c.comps) - 1)) )
            j += 1

            #
            # After a run of scaffold components there should be a gap separating the next scaffold.
            #
            if j < max_index:
                if components[j].type != "N":
                    print("Error: gap not found where expected, chromosome " + components[j].chr + ", component " + str(j) + ".\n", file=sys.stderr)
                    sys.exit()
                chr_comp = ChrComp(chr, "gap", "gap", components[j].chr_sbp, components[j].chr_ebp)
                c.comps.append(chr_comp)
                last_scaffold = components[j+1].scaf
            i = j + 1

        chromosomes[c.chr] = c

def merge_agp(chromosomes, chromosomes_key, chr_agp, scaf_agp, scaf_agp_key):
    #
    # Here we assume a single AGP file has been specified describing the scaffolds within each chromosome.
    #
    for chr in chr_agp:
        #
        # Create chromosome object.
        #
        c = Chr(chr)

        for k in range(0, len(chr_agp[chr])):
            #
            # Create chromosome component object to hold scaffolds/gaps.
            #
            if chr_agp[chr][k].type != "N":
                chr_comp = ChrComp(chr, "scaffold", chr_agp[chr][k].sub, chr_agp[chr][k].start, chr_agp[chr][k].end)
                chr_comp.ori = chr_agp[chr][k].sub_ori

                cc = CtgComp(chr, chr_agp[chr][k].type, chr_agp[chr][k].part, chr_agp[chr][k].sub_sbp, chr_agp[chr][k].end)
                cc.name = chr_agp[chr][k].sub
                cc.sbp  = chr_agp[chr][k].sub_sbp
                cc.ebp  = chr_agp[chr][k].sub_ebp
            else:
                chr_comp = ChrComp(chr, "gap", chr_agp[chr][k].sub, chr_agp[chr][k].start, chr_agp[chr][k].end)
                cc = CtgComp(chr, chr_agp[chr][k].type, chr_agp[chr][k].part, chr_agp[chr][k].start, chr_agp[chr][k].end)
            chr_comp.comps.append(cc)
            #
            # Store this scaffold object and where to find it.
            #
            c.comps.append(chr_comp)
            c.key[chr_comp.name] = len(chr_comp.comps) - 1
            if chr_comp.name != "gap":
                #
                # For contigs/scaffolds at the chromosome level, the names may not be unique across the genome,
                # but must be unique within a chromosome, so a contig may be found in more than one chromosome, so
                # store all locations within an array. 
                #
                if chr_comp.name not in chromosomes_key:
                    chromosomes_key[chr_comp.name] = []
                chromosomes_key[chr_comp.name].append( (chr, (len(c.comps) - 1)) )

        chromosomes[c.chr] = c

    
def merge_ann(chromosomes, ann):
    global synthetic_chr

    print("  Merging annotation and genome structure...", file=sys.stderr, end="", flush=True)
    
    #
    # Identify which scaffolds annotated objects belong to (aka, which scaffolds contain which genes)
    #
    for chr in chromosomes:
        #
        # Find the maximum basepair on this chromosome.
        #
        max_bp = 0
        for k in range(0, len(chromosomes[chr].comps)):
            if chromosomes[chr].comps[k].end > max_bp:
                max_bp = chromosomes[chr].comps[k].end
        #
        # Sort the gene objects for this chromosome by starting basepair.
        #
        genes = []
        for gene in ann:
            if ann[gene][0].chr == chr:
                # Find the minimum and maximum basepairs for this gene.
                start = max_bp
                end   = 0
                for comp in ann[gene]:
                    if comp.sbp < start:
                        start = comp.sbp
                    if comp.ebp > end:
                        end = comp.ebp
                genes.append( (gene, start, end) )

        if len(genes) == 0:
            continue

        genes.sort(key=lambda gene: gene[1])
        gene_index = 0

        #
        # Co-iterate over the set of scaffolds and the set of genes to assign genes to their encapsulating scaffolds.
        #
        for k in range(0, len(chromosomes[chr].comps)):
            if chromosomes[chr].comps[k].type == "gap":
                continue
            scaf_sbp = chromosomes[chr].comps[k].start
            scaf_ebp = chromosomes[chr].comps[k].end

            # print("Scaffold " + chromosomes[chr].comps[k].name + ", start:" + str(scaf_sbp) + "; end: " + str(scaf_ebp), file=sys.stderr)
            
            # Gene crosses the border of the scaffold.
            while gene_index < len(genes) and (genes[gene_index][1] < scaf_sbp and genes[gene_index][2] > scaf_sbp):
                chromosomes[chr].genes.append(genes[gene_index][0])
                # print("    Crosses border gene: " + str(genes[gene_index][0]) + ", start: " + str(genes[gene_index][1]) + "; end: " + str(genes[gene_index][2]), file=sys.stderr)
                gene_index += 1
            # Gene is contained within the scaffold.
            while gene_index < len(genes) and genes[gene_index][1] >= scaf_sbp and genes[gene_index][1] <= scaf_ebp:
                chromosomes[chr].comps[k].genes.append(genes[gene_index][0])
                # print("    Contained gene: " + str(genes[gene_index][0]) + ", start: " + str(genes[gene_index][1]) + "; end: " + str(genes[gene_index][2]), file=sys.stderr)
                gene_index += 1
            # Gene crosses the other border of the scaffold.
            while gene_index < len(genes) and (genes[gene_index][1] < scaf_ebp and genes[gene_index][2] > scaf_ebp):
                chromosomes[chr].genes.append(genes[gene_index][0])
                # print("    Crosses border gene: " + genes[gene_index][0] + ", start: " + str(genes[gene_index][1]) + "; end: " + str(genes[gene_index][2]), file=sys.stderr)
                gene_index += 1

    #
    # We need to handle unordered scaffolds specially. 
    #
    if unordered_scaffolds_concatenated:
        for k in range(0, len(chromosomes[synthetic_chr].comps)):
            if chromosomes[synthetic_chr].comps[k].type == "gap":
                continue
            scaf     = chromosomes[synthetic_chr].comps[k].name
            scaf_sbp = chromosomes[synthetic_chr].comps[k].start
            scaf_ebp = chromosomes[synthetic_chr].comps[k].end

            genes = []
            for gene in ann:
                if ann[gene][0].chr == scaf:
                    # Find the minimum and maximum basepairs for this gene.
                    start = scaf_ebp
                    end   = 0
                    for comp in ann[gene]:
                        if comp.sbp < start:
                            start = comp.sbp
                        if comp.ebp > end:
                            end = comp.ebp
                    genes.append( (gene, start, end) )
                    if len(genes) == 0:
                        continue

            genes.sort(key=lambda gene: gene[1])
            gene_index = 0

            while gene_index < len(genes):
                chromosomes[synthetic_chr].comps[k].genes.append(genes[gene_index][0])
                gene_index += 1

    print("done.", file=sys.stderr)


def merge_fasta(chromosomes, fasta):
    #
    # Fetch the subsequence for each scaffold from the FASTA file.
    #
    for chr in chromosomes:
        if chr not in fasta:
            print("Warning: unable to find FASTA sequence for '" + chr + "'", file=sys.stderr)
            continue
        for k in range(0, len(chromosomes[chr].comps)):
            if chromosomes[chr].comps[k].type == "gap":
                continue
            chromosomes[chr].comps[k].seq = fasta[chr][chromosomes[chr].comps[k].start-1:chromosomes[chr].comps[k].end]
            # print chromosomes[chr].comps[k].name, "start", chromosomes[chr].comps[k].start-1, "end", chromosomes[chr].comps[k].end-1, file=sys.stderr)


def parse_changes_file(changes_path, changes):
    global defined_operations

    #
    # Load a changes file defining how to change the assembly.
    # Chromosome    source_scaffold/chromosome     Operation     destination_scaffold/landmark     orientation
    #
    fh = open(changes_path)

    lnum  = 0
    for line in fh:
        lnum += 1
        line  = line.strip('\n ')

        if len(line) == 0 or line[0] == "#":
            continue

        parts = line.split('\t')

        if len(parts) < 3:
            print("Error: malformed line, not enough columns for line " + str(lnum) + " of " + changes_path, file=sys.stderr)
            sys.exit()

        if parts[2] not in defined_operations:
            print("Error: undefined operation '" + parts[2] + "' on line " + str(lnum) + " of " + changes_path, file=sys.stderr)
            sys.exit()

        if parts[2] == "invert" or parts[2] == "invert_scaffold" or parts[2] == "rename":
            op = Op(parts[0], parts[1], parts[2])
        else:
            if len(parts) < 5:
                print("Error: malformed line, not enough columns for line " + str(lnum) + " of " + changes_path, file=sys.stderr)
                sys.exit()
            op = Op(parts[0], parts[1], parts[2], parts[3], parts[4])

        # if op.chr not in changes:
        #     changes[op.chr] = []
        changes.append(op)

    fh.close()

    cnt = set()
    for ch in changes:
         cnt.add(ch.chr)
    print("Loaded", len(changes), "operations for", len(cnt), "chromosomes.", file=sys.stderr)


def find_scaffold_location(chrs_key, op):
    chrom = ""
    index = -1

    if op.source in chrs_key and op.source.lower() != "null":
        for tup in chrs_key[op.source]:
            if tup[0] == op.chr:
                (chrom, index) = tup
                break
        if len(chrom) == 0:
            print("Error: Unable to find scaffold '{}' on chromosome '{}'".format(op.source, op.chr), file=sys.stderr)
            sys.exit()
        op.chr_src   = chrom
        op.index_src = index
    else:
        if len(op.source) > 0 and len(op.dest) == 0:
            print("Error: Unable to find component, '{}', it is not defined in genome structure.".format(op.source), file=sys.stderr)
            exit(1)
        
    chrom = ""
    index = -1
    
    if op.dest in chrs_key and op.dest.lower() != "null":
        for tup in chrs_key[op.dest]:
            if tup[0] == op.chr:
                (chrom, index) = tup
                break
        if len(chrom) == 0:
            print("Error: Unable to find scaffold '{}' on chromosome '{}'".format(op.dest, op.chr), file=sys.stderr)
            sys.exit()
        op.chr_dest   = chrom
        op.index_dest = index
    else:
        if len(op.dest) > 0:
            print("Error: Unable to find component, '{}'; it is not defined in genome strucutre.".format(op.dest), file=sys.stderr)
            exit(1)

    return

def invert_chrcomp(scaf, ann, fasta, new_scaf_beg_bp, new_scaf_end_bp):
    #
    # Reverse the list of contig objects.
    #
    scaf.comps.reverse()

    #
    # Adjust the orientation.
    #
    if scaf.ori == "+":
        scaf.ori = "-"
    else:
        scaf.ori = "+"

    #
    # Adjust the internal coordinates and renumber the components.
    #
    bp   = scaf.start
    part = 1
    for i in range(0, len(scaf.comps)):
        scaf.comps[i].chr_sbp = bp
        scaf.comps[i].chr_ebp = bp + scaf.comps[i].len - 1
        scaf.comps[i].part    = part
        bp   += scaf.comps[i].len
        part += 1

    #
    # Extract the sequence and reverse complement it.
    #
    if len(fasta) > 0:
        scaf.seq = reverse_complement(scaf.seq)

    #
    # Adjust the coordinates of gene components.
    #
    for i in range(0, len(scaf.genes)):
        gcomps = ann[scaf.genes[i]]

        # print("      Gene " + scaf.genes[i] + "; start: " + str(gcomps[0].sbp) + "; end: " + str(gcomps[0].ebp) + "; len: " + str(gcomps[0].ebp - gcomps[0].sbp + 1))

        for j in range(0, len(gcomps)):
            dist_from_beg = gcomps[j].sbp - scaf.start
            new_end_bp = new_scaf_end_bp - dist_from_beg
            new_beg_bp = new_end_bp - (gcomps[j].ebp - gcomps[j].sbp)

            gcomps[j].sbp = new_beg_bp
            gcomps[j].ebp = new_end_bp
            if gcomps[j].strand == "+":
                gcomps[j].strand = "-"
            else:
                gcomps[j].strand = "+"

        # print("      Adjusted Gene " + scaf.genes[i] + "; start: " + str(gcomps[0].sbp) + "; end: " + str(gcomps[0].ebp) + "; len: " + str(gcomps[0].ebp - gcomps[0].sbp + 1))


def invert_scaffold(chrs, chrs_key, op, ann, fasta):
    #
    # Find the scaffold to invert.
    #
    find_scaffold_location(chrs_key, op)
    (chr_src, index_src) = (op.chr_src, op.index_src)

    print("Scaffold to invert {} is on chr {}, index {}".format(op.source, chr_src, index_src), file=sys.stderr)
    
    scaf = chrs[chr_src].comps[index_src]

    invert_chrcomp(scaf, ann, fasta, scaf.start, scaf.end)


def remove_scaffold(op, chrs_key, chrs, gff, new_comp_list):
    global synthetic_chr

    find_scaffold_location(chrs_key, op)
    (chr_src, index_src) = (op.chr_src,  op.index_src)
    
    comp_list = chrs[chr_src].comps
    
    index = 0
    while index < index_src:
        new_comp_list.append(comp_list[index])
        index += 1
    scaffold = comp_list[index_src]
    index    = index_src + 1

    #
    # Remove the gap immediately after it; if this is the final scaffold, remove the preceeding gap.
    #
    gap_len = 0
    if index_src < len(comp_list) - 1:
        gap_len = comp_list[index].len
        index += 1
    elif index_src == len(comp_list) - 1:
        gap_len = comp_list[index_src - 1].len

    #
    # Adjust the coordinates for any annotated genes on this scaffold.
    #
    for gene in scaffold.genes:
        for i in range(0, len(gff[gene])):
            g = gff[gene][i]
            if chr_src != synthetic_chr:
                g.sbp = g.sbp - scaffold.start
                g.ebp = g.ebp - scaffold.start
            g.chr = scaffold.name

    while index < len(comp_list):
        #
        # Adjust the coordinates downstream.
        #
        for i in range(0, len(comp_list[index].comps)):
            comp_list[index].comps[i].chr_sbp = comp_list[index].comps[i].chr_sbp - scaffold.len - gap_len
            comp_list[index].comps[i].chr_ebp = comp_list[index].comps[i].chr_sbp - scaffold.len - gap_len
        comp_list[index].start = comp_list[index].start - scaffold.len - gap_len
        comp_list[index].end   = comp_list[index].end   - scaffold.len - gap_len
        # Record the adjusted scaffold into the new list.
        new_comp_list.append(comp_list[index])
        # Update the chromosomes key as to where to find this component.
        chrs_key[comp_list[index].name] = (chr_src, (len(new_comp_list) - 1))        
        # Update any gene coordinates.
        if chr_src != synthetic_chr:
            for gene in comp_list[index].genes:
                for i in range(0, len(gff[gene])):
                    gff[gene][i].sbp = gff[gene][i].sbp - scaffold.len - gap_len
                    gff[gene][i].ebp = gff[gene][i].ebp - scaffold.len - gap_len
        index += 1

    #
    # If the removal of a scaffold was at the back of the list and left a gap behind, remove it.
    #
    if index_src == len(comp_list) - 1:
        new_comp_list.pop()

    return scaffold

def insert_scaffold_before(chrs, chrs_key, op, gff, fasta):
    global scaf_gap_size

    #
    # Find the locations of the scaffold to insert, and the scaffold we will insert before.
    #
    find_scaffold_location(chrs_key, op)
    (chr_src,  index_src)  = (op.chr_src,  op.index_src)
    (chr_dest, index_dest) = (op.chr_dest, op.index_dest)

    print("  Scaffold to insert {} is on chr {}, index {}".format(op.source, chr_src, index_src), file=sys.stderr)
    print("  Scaffold to insert before {} is on chr {}, index {}".format(op.dest, chr_dest, index_dest), file=sys.stderr)

    new_comp_list = []
    #
    # First, remove the scaffold from its old location.
    #
    scaffold = remove_scaffold(op, chrs_key, chrs, gff, new_comp_list)

    (chr_src, index_src)   = chrs_key[op.source]
    (chr_dest, index_dest) = chrs_key[op.dest]
    chrs[chr_src].comps    = new_comp_list

    #
    # Check if we have sequence for this scaffold. If not, look for it in the fasta dictionary.
    #
    if len(fasta) > 0 and len(scaffold.seq) == 0:
        if scaffold.name not in fasta:
            print("Error: Unable to find the sequence for " + scaffold.name + ".", file=sys.stderr)
            sys.exit()
        else:
            scaffold.seq = fasta[scaffold.name]

    #
    # If requested, reverse complement the sequence.
    #
    if op.ori == "-":
        scaffold.seq = reverse_complement(scaffold.seq)
        #
        # Adjust the orientation.
        #
        if scaffold.ori == "+":
            scaffold.ori = "-"
        else:
            scaffold.ori = "+"


    #
    # Create a new list of components describing this chromosome, in the new order.
    #
    comp_list     = chrs[chr_dest].comps
    new_comp_list = []

    index = 0
    while index < index_dest:
        new_comp_list.append(comp_list[index])
        index += 1

    #
    # Add the moved scaffold.
    #
    scaffold.start = comp_list[index - 1].end + 1 if index_dest > 0 else 1
    scaffold.end   = scaffold.start + scaffold.len - 1
    for i in range(0, len(scaffold.comps)):
        if i == 0:
            scaffold.comps[i].chr_sbp = scaffold.start
        else:
            scaffold.comps[i].chr_sbp = scaffold.comps[i - 1].chr_ebp + 1
        scaffold.comps[i].chr_ebp = scaffold.comps[i].chr_sbp + scaffold.comps[i].len - 1
    new_comp_list.append(scaffold)
    #
    # Update the chromosome key on where to find this scaffold.
    #
    chrs_key[scaffold.name] = (chr_dest, (len(new_comp_list) - 1))

    #
    # Adjust the coordinates for any annotated genes on this scaffold.
    #
    for gene in scaffold.genes:
        for i in range(0, len(gff[gene])):
            g     = gff[gene][i]
            g.chr = chr_dest

            if op.ori == "-":
                prev_sbp = g.sbp
                prev_ebp = g.ebp
                g.sbp = scaffold.start + scaffold.len - prev_sbp
                g.ebp = scaffold.start + scaffold.len - (prev_ebp - prev_sbp + 1)
            else:
                g.sbp = g.sbp + scaffold.start - 1
                g.ebp = g.ebp + scaffold.start - 1

    #
    # Add a new gap to space the moved scaffold.
    #
    new_start = scaffold.end + 1
    new_end   = new_start + scaf_gap_size - 1
    new_comp_list.append(ChrComp(chr, "gap", "gap", new_start, new_end))

    while index < len(comp_list):
        #
        # Adjust the coordinates downstream.
        #
        for i in range(0, len(comp_list[index].comps)):
            comp_list[index].comps[i].chr_sbp = comp_list[index].comps[i].chr_sbp + scaffold.len + scaf_gap_size
            comp_list[index].comps[i].chr_ebp = comp_list[index].comps[i].chr_ebp + scaffold.len + scaf_gap_size
        comp_list[index].start = comp_list[index].start + scaffold.len + scaf_gap_size
        comp_list[index].end   = comp_list[index].end   + scaffold.len + scaf_gap_size
        # Record the adjusted scaffold into the new list.
        new_comp_list.append(comp_list[index])
        # Update the chromosomes key as to where to find the moved components.
        chrs_key[comp_list[index].name] = (chr_dest, (len(new_comp_list) - 1))
        # Update any gene coordinates.
        for gene in comp_list[index].genes:
            for i in range(0, len(gff[gene])):
                gff[gene][i].sbp = gff[gene][i].sbp + scaffold.len + scaf_gap_size
                gff[gene][i].ebp = gff[gene][i].ebp + scaffold.len + scaf_gap_size
        index += 1

    chrs[chr_dest].comps = new_comp_list

def insert_scaffold_after(chrs, chrs_key, op, gff, fasta):
    global scaf_gap_size

    #
    # Find the locations of the scaffold to insert, and the scaffold we will insert after.
    #
    find_scaffold_location(chrs_key, op)
    (chr_src,  index_src)  = (op.chr_src,  op.index_src)
    (chr_dest, index_dest) = (op.chr_dest, op.index_dest)

    print("  Scaffold to insert {} is on chr {}, index {}".format(op.source, chr_src, index_src), file=sys.stderr)
    print("  Scaffold to insert after {} is on chr {}, index {}".format(op.dest, chr_dest, index_dest), file=sys.stderr)

    new_comp_list = []
    #
    # First, remove the scaffold from its old location.
    #
    scaffold = remove_scaffold(op, chrs_key, chrs, gff, new_comp_list)
    # @ARC: I added the `[0]` since the object returned is a list of tuples
    # of length 1,  [(chr, idx)]
    (chr_src, index_src)   = chrs_key[op.source][0]
    (chr_dest, index_dest) = chrs_key[op.dest][0]
    chrs[chr_src].comps    = new_comp_list

    #
    # Check if we have sequence for this scaffold. If not, look for it in the fasta dictionary.
    #
    if len(fasta) > 0 and len(scaffold.seq) == 0:
        if scaffold.name not in fasta:
            print("Error: Unable to find the sequence for {}.".format(scaffold.name), file=sys.stderr)
            sys.exit()
        else:
            scaffold.seq = fasta[scaffold.name]

    #
    # If requested, reverse complement the sequence.
    #
    if op.ori == "-":
        scaffold.seq = reverse_complement(scaffold.seq)
        #
        # Adjust the orientation.
        #
        if scaffold.ori == "+":
            scaffold.ori = "-"
        else:
            scaffold.ori = "+"

    #
    # Create a new list of components describing this chromosome, in the new order.
    #
    comp_list     = chrs[chr_dest].comps
    new_comp_list = []

    index = 0
    while index <= index_dest:
        new_comp_list.append(comp_list[index])
        index += 1
    #
    # Add a new gap to space the moved scaffold.
    #
    new_start = comp_list[index_dest].end + 1
    new_end   = new_start + scaf_gap_size - 1
    new_comp_list.append(ChrComp(chr, "gap", "gap", new_start, new_end))
    #
    # Add the moved scaffold.
    #
    scaffold.start = new_end + 1
    scaffold.end   = scaffold.start + scaffold.len - 1
    for i in range(0, len(scaffold.comps)):
        if i == 0:
            scaffold.comps[i].chr_sbp = scaffold.start
        else:
            scaffold.comps[i].chr_sbp = scaffold.comps[i - 1].chr_ebp + 1
        scaffold.comps[i].chr_ebp = scaffold.comps[i].chr_sbp + scaffold.comps[i].len - 1
    new_comp_list.append(scaffold)
    #
    # Update the chromosome key on where to find this scaffold.
    #
    # @ARC: added the brackets to mantain the list of tuples structure
    chrs_key[scaffold.name] = [(chr_dest, (len(new_comp_list) - 1))]

    #
    # Adjust the coordinates for any annotated genes on this scaffold.
    #
    for gene in scaffold.genes:
        for i in range(0, len(gff[gene])):
            g     = gff[gene][i]
            g.chr = chr_dest

            if op.ori == "-":
                prev_sbp = g.sbp
                prev_ebp = g.ebp
                g.sbp = scaffold.start + scaffold.len - prev_sbp
                g.ebp = scaffold.start + scaffold.len - (prev_ebp - prev_sbp + 1)
            else:
                g.sbp = g.sbp + scaffold.start - 1
                g.ebp = g.ebp + scaffold.start - 1

    while index < len(comp_list):
        #
        # Adjust the coordinates downstream.
        #
        comp_list[index].start = comp_list[index].start + scaffold.len + scaf_gap_size
        comp_list[index].end   = comp_list[index].end   + scaffold.len + scaf_gap_size
        for i in range(0, len(comp_list[index].comps)):
            comp_list[index].comps[i].chr_sbp = comp_list[index].comps[i].chr_sbp + scaffold.len + scaf_gap_size
            comp_list[index].comps[i].chr_ebp = comp_list[index].comps[i].chr_ebp + scaffold.len + scaf_gap_size
        # Record the adjusted scaffold into the new list.
        new_comp_list.append(comp_list[index])
        # Update the chromosomes key as to where to find the moved components.
        # @ARC: added the brackets to mantain the list of tuples structure
        chrs_key[comp_list[index].name] = [(chr_dest, (len(new_comp_list) - 1))]
        # Update any gene coordinates.
        for gene in comp_list[index].genes:
            for i in range(0, len(gff[gene])):
                gff[gene][i].sbp = gff[gene][i].sbp + scaffold.len + scaf_gap_size
                gff[gene][i].ebp = gff[gene][i].ebp + scaffold.len + scaf_gap_size

        index += 1

    chrs[chr_dest].comps = new_comp_list


def split_chromosome_after(chrs, chrs_key, op, ann, fasta):
    global scaf_gap_size

    #
    # Check that the new chromosome doesn't already exist.
    #
    if op.source in chrs_key:
        print("Error: new chromosome '{}' is already defined in the assembly.".format(op.source), file=sys.stderr)
        sys.exit()

    if op.dest not in chrs_key:
        print("Error: unable to find scaffold '{}' on chromosome {}.".format(op.dest, op.chr), file=sys.stderr)
        sys.exit()
        
    #
    # Find the scaffold that marks the chromosome split point.
    #
    find_scaffold_location(chrs_key, op)
    (chr_src, index_src)  = (op.chr_dest,  op.index_dest)

    print("  Found '{}' that marks split of chromosome '{}' at index {}. New chromosome will be named '{}'".format(
        op.dest, chr_src, index_src + 1, op.source), file=sys.stderr)

    #
    # Create the new chromosome.
    #
    c = Chr(op.source)
    
    src_comp_list = []
    new_comp_list = []

    #
    # Move the components we are keeping on the old chromosome to src_comp_list.
    #
    comp_list = chrs[chr_src].comps

    index = 0
    while index <= index_src:
        src_comp_list.append(comp_list[index])
        index += 1

    #
    # There should be a gap immediately after index_src that we will remove.
    #
    if comp_list[index_src + 1].type != "gap":
        print("Error: gap not found prior to scaffold marking split; found '{}' at position {}".format(comp_list[index_src + 1].type, index_src+1), file=sys.stderr)
        sys.exit()

    index = index_src + 2
    while index < len(comp_list):
        new_comp_list.append(comp_list[index])
        index += 1

    #
    # Place the split components back into their respective chromosomes.
    #
    print("  Source chromosome reduced from {} to {} components. New chromosome contains {} components.".format(
        len(chrs[chr_src].comps), len(src_comp_list), len(new_comp_list)), file=sys.stderr)
    chrs[chr_src].comps = src_comp_list
    c.comps = new_comp_list

    #
    # Adjust the chromosome coordinates for the new chromosome components.
    #
    bp_adj   = c.comps[0].start - 1
    obj_cnt  = 0
    gene_cnt = 0
    for i in range(len(c.comps)):
        c.comps[i].chr   = c.chr
        c.comps[i].start = c.comps[i].start - bp_adj
        c.comps[i].end   = c.comps[i].end   - bp_adj
    
        #
        # Adjust the coordinates for any annotated genes on this scaffold.
        #
        scaffold = c.comps[i]
        for gene in scaffold.genes:
            gene_cnt += 1
            for j in range(0, len(ann[gene])):
                g     = ann[gene][j]
                g.chr = c.chr
                g.sbp = g.sbp - bp_adj
                g.ebp = g.ebp - bp_adj
                obj_cnt  += 1

    print("  Moved {} genes, composed of {} annotation objects, to new chromosome.".format(gene_cnt, obj_cnt), file=sys.stderr)
    
    #
    # Update the chromosome key on where to find all moved scaffolds.
    #
    for i in range(len(c.comps)):
        comp_name = c.comps[i].name
        if comp_name in chrs_key:
            for j in range(len(chrs_key[comp_name])):
                if chrs_key[comp_name][j][0] == op.chr:
                    chrs_key[comp_name][j] = (op.source, i)
                    break

    chrs[c.chr] = c

    return

def fuse_chromosome(chrs, chrs_key, op, ann, fasta):
    global scaf_gap_size

    #
    # Check that the chromosome to fuse exists.
    #
    if op.source not in chrs:
        print("Error: chromosome to fuse, '{}' is not defined in the assembly.".format(op.source), file=sys.stderr)
        sys.exit()

    chr_src    = op.source
    index_src  = len(chrs[chr_src].comps)
    chr_dest   = op.chr
    index_dest = len(chrs[chr_dest].comps)

    print("  Chromosome/scaffold '{}' will be fused onto the end of chromosome/scaffold '{}' in orientation '{}'.".format(
        chr_src, chr_dest, op.ori), file=sys.stderr)

    #
    # Get the end basepair coordinate of chr_dest (where we will append the fused components).
    #
    chr_dest_end_bp = chrs[chr_dest].comps[index_dest - 1].end
    c               = chrs[chr_src]
    
    new_gap_start = chr_dest_end_bp + 1
    new_gap_end   = new_gap_start + scaf_gap_size - 1
    new_comp_no   = index_dest + 1

    #
    # Add a new gap to the end of chr_dest to space the fused scaffold/chromosome.
    #
    ctgcmp = CtgComp(op.chr, "gap", new_comp_no, new_gap_start, new_gap_end)
    chrcmp = ChrComp(op.chr, "gap", "gap", new_gap_start, new_gap_end)
    chrcmp.comps.append(ctgcmp)
    chrs[chr_dest].comps.append(chrcmp)

    #
    # If requested, invert the scaffold and gene components before fusion.
    #
    if op.ori == "-":
        chr_len = c.comps[-1].end
        end_bp  = chr_len

        c.comps.reverse()
        reordered_comps = []
        while len(c.comps) > 0:
            cc      = c.comps.pop()
            new_end = end_bp
            new_beg = end_bp - cc.len + 1
            end_bp  = new_beg - 1
            #
            # Adjust the internal components including genes, contigs and sequence.
            #
            invert_chrcomp(cc, ann, fasta, new_beg, new_end)
        
            cc.start = new_beg
            cc.end   = new_end
        
            reordered_comps.append(cc)

        #
        # Reverse the components list to be in the proper order once again.
        #
        reordered_comps.reverse()
        c.comps = reordered_comps

    #
    # Update the chromosome key on where to find the moved components of the fused scaffold/chromosome.
    #
    for i in range(len(c.comps)):
        comp_name = c.comps[i].name
        if comp_name in chrs_key:
            for j in range(len(chrs_key[comp_name])):
                if chrs_key[comp_name][j][0] == op.source:
                    chrs_key[comp_name][j] = (op.chr, new_comp_no + i)
                    break
    new_comp_no += 1


    #
    # Adjust the chromosome coordinates for the new chromosome components and move them to the fused chromosome.
    #
    bp_adj   = new_gap_end
    obj_cnt  = 0
    gene_cnt = 0
    cmp_cnt  = 0
    for i in range(len(c.comps)):
        c.comps[i].chr   = chr_dest
        c.comps[i].start = c.comps[i].start + bp_adj
        c.comps[i].end   = c.comps[i].end   + bp_adj
        cmp_cnt += 1

        #
        # Adjust the coordinates for any annotated genes on this scaffold.
        #
        scaffold = c.comps[i]
        for gene in scaffold.genes:
            gene_cnt += 1
            for j in range(0, len(ann[gene])):
                g     = ann[gene][j]
                g.chr = chr_dest
                g.sbp = g.sbp + bp_adj
                g.ebp = g.ebp + bp_adj
                obj_cnt  += 1

        chrs[chr_dest].comps.append(c.comps[i])

    print("  Moved {} chromosome components and {} genes, composed of {} annotation objects, to new chromosome.".format(cmp_cnt, gene_cnt, obj_cnt), file=sys.stderr)

    del chrs[chr_src]
    return

def invert_chromosome(chrs, chrs_key, op, ann, fasta):

    #
    # Check that the chromosome to invert exists.
    #
    if op.chr not in chrs:
        print("Error: chromosome to fuse, '{}' is not defined in the assembly.".format(op.chr), file=sys.stderr)
        sys.exit()

    chr_src = op.chr
    c       = chrs[chr_src]

    #
    # Treat the list of chromosome components as a FIFO queue. Reverse the list so we can
    # efficiently pop() each element off the end for reordering.
    #
    comp_cnt = 0
    gene_cnt = 0
    chr_len  = c.comps[-1].end
    end_bp   = chr_len

    c.comps.reverse()
    reordered_comps = []
    while len(c.comps) > 0:
        cc      = c.comps.pop()
        new_end = end_bp
        new_beg = end_bp - cc.len + 1
        end_bp  = new_beg - 1
        #
        # Adjust the internal components including genes, contigs and sequence.
        #
        invert_chrcomp(cc, ann, fasta, new_beg, new_end)
        
        cc.start = new_beg
        cc.end   = new_end
        
        reordered_comps.append(cc)
        comp_cnt += 1
        gene_cnt += len(cc.genes)


    #
    # Reverse the components list to be in the proper order once again.
    #
    reordered_comps.reverse()

    c.comps = reordered_comps

    #
    # Update the chromosome key on where to find the moved components of the fused scaffold/chromosome.
    #
    for i in range(len(c.comps)):
        comp_name = c.comps[i].name
        if comp_name in chrs_key:
            for j in range(len(chrs_key[comp_name])):
                if chrs_key[comp_name][j][0] == chr_src:
                    chrs_key[comp_name][j] = (op.chr, i)
                    break

    print("  Inverted {} components containing {} genes.".format(comp_cnt, gene_cnt), file=sys.stderr)

    return

def rename(chrs, chrs_key, op, ann, fasta):

    chr_src  = op.chr
    chr_dest = op.source
    c       = chrs[chr_src]
    c.chr   = chr_dest

    #
    # Check that chr_dest does not already occur in the dataset.
    #
    if chr_dest in chrs:
        print("  Error: '{}' is already defined in the set of chromosomes.".format(chr_dest), file=sys.stderr)
        sys.exit()

    agp_cnt  = 0
    gene_cnt = 0
    obj_cnt  = 0
    #
    # Update the chromosome key on where to find the moved components of the fused scaffold/chromosome.
    #
    for i in range(len(c.comps)):
        comp_name = c.comps[i].name
        if comp_name in chrs_key:
            for j in range(len(chrs_key[comp_name])):
                if chrs_key[comp_name][j][0] == chr_src:
                    chrs_key[comp_name][j] = (chr_dest, chrs_key[comp_name][j][1])
                    agp_cnt += 1
                    break
        #
        # Adjust the chromosome name for any annotated genes/objects.
        #
        for gene in c.comps[i].genes:
            gene_cnt += 1
            for j in range(0, len(ann[gene])):
                g     = ann[gene][j]
                g.chr = chr_dest
                obj_cnt  += 1


    chrs[chr_dest] = chrs[chr_src]
    del chrs[chr_src]

    print("  Renamed {} assembly components; {} genes, composed of {} annotation objects, to reside on chromosome '{}'.".format(agp_cnt, gene_cnt, obj_cnt, chr_dest), file=sys.stderr)
    return


def reverse_complement(seq):
    map = {'A': 'T',
           'a': 'T',
           'C': 'G',
	   'c': 'G',
           'G': 'C',
           'g': 'C',
           'T': 'A',
           't': 'A',
           'N': 'N',
           'n': 'N',
           'W': 'W',
           'w': 'W',
           'S': 'S',
           's': 'S',
           'M': 'K',
           'm': 'K',
           'K': 'M',
           'k': 'M',
           'R': 'Y',
           'r': 'Y',
           'Y': 'R',
           'y': 'R'}
    slen  = len(seq)
    i     = slen - 1
    rcseq = ""
    while i >= 0:
        if seq[i] not in map:
            rcseq += 'N'
        else:
            rcseq += map[seq[i]]
        i -= 1
    return rcseq

def write_agp(agp_path, chromosomes, datestamp):
    global out_path, out_prefix

    suff = False
    gzip = False
    if agp_path[-7:] == ".agp.gz":
        suff = True
        gzip = True
    elif agp_path[-4:] == ".agp":
        suff = True
    elif agp_path[-3:] == ".gz":
        gzip = True

    if len(out_prefix) > 0:
        path = out_path + out_prefix
        suff = True
    else:
        pos  = agp_path.rfind('/')
        if suff and gzip:
            path = out_path + agp_path[pos:-7]
        elif suff:
            path = out_path + agp_path[pos:-4]
        elif gzip:
            path = out_path + agp_path[pos:-3]
        else:
            path = out_path + agp_path[pos:]
        path += "." + datestamp

    if suff and gzip:
        scaf_out_path = path + ".agp.gz"
        ctg_out_path  = path + ".contig.agp.gz"
    elif gzip:
        scaf_out_path = path + ".gz"
        ctg_out_path  = path + ".contig.gz"
    elif suff:
        scaf_out_path = path + ".agp"
        ctg_out_path  = path + ".contig.agp"

    if gzip:
        sout_fh = gzip.open(scaf_out_path, "wb")
        if len(ctg_agp_path) > 0:
            cout_fh = gzip.open(ctg_out_path, "wb")
    else:
        sout_fh = open(scaf_out_path, 'w')
        if len(ctg_agp_path) > 0:
            cout_fh = open(ctg_out_path, 'w')

    first = True
    for chr in chromosomes:
        if first == False:
            sout_fh.write("###\n")
        comps = chromosomes[chr].comps
        for i in range(0, len(comps)):
            if comps[i].type == "scaffold":
                sout_fh.write(chr + "\t" + str(comps[i].start) + "\t" + str(comps[i].end) + "\t" + str(i+1) + "\t" +
                              "W" + "\t" + comps[i].name + "\t" + "1" + "\t" + str(comps[i].len) + "\t" + comps[i].ori + "\n")
            else:
                sout_fh.write(chr + "\t" + str(comps[i].start) + "\t" + str(comps[i].end) + "\t" + str(i+1) + "\t" +
                              "N" + "\t" + str(comps[i].len) + "\tfragment\tyes\n")
        first = False
    sout_fh.close()

    if len(ctg_agp_path) > 0:
        first = True
        for chr in chromosomes:
            if first == False:
                cout_fh.write("###\n")
                partno = 1
                comps  = chromosomes[chr].comps
            for i in range(0, len(comps)):
                if comps[i].type == "scaffold":
                    for j in range(0, len(comps[i].comps)):
                        if comps[i].comps[j].type == "W":
                            cout_fh.write(chr + "\t" + str(comps[i].comps[j].chr_sbp) + "\t" + str(comps[i].comps[j].chr_ebp) + "\t" + str(partno) + "\t" +
                                          "W" + "\t" + comps[i].comps[j].ctg + "\t" + "1" + "\t" + str(comps[i].comps[j].len) + "\t" + comps[i].ori + "\n")
                        else:
                            cout_fh.write(chr + "\t" + str(comps[i].comps[j].chr_sbp) + "\t" + str(comps[i].comps[j].chr_ebp) + "\t" + str(partno) + "\t" +
                                          "N" + "\t" + str(comps[i].comps[j].len) + "\tfragment\tyes\n")
                            partno += 1
                else:
                    cout_fh.write(chr + "\t" + str(comps[i].start) + "\t" + str(comps[i].end) + "\t" + str(partno) + "\t" +
                                  "N" + "\t" + str(comps[i].len) + "\tfragment\tyes\n")
            partno += 1
            first = False
        cout_fh.close()

def write_gff(gff_path, chromosomes, gff, datestamp):
    global out_path

    suff = False
    lsuf = False
    gzip = False
    if gff_path[-7:] == ".gff.gz":
        suff = True
        gzip = True
    elif gff_path[-8:] == ".gff3.gz":
        suff = True
        gzip = True
        lsuf = True
    elif gff_path[-4:] == ".gff":
        suff = True
    elif gff_path[-4:] == ".gff3":
        suff = True
        lsuf = True
    elif gff_path[-3:] == ".gz":
        gzip = True

    if len(out_prefix) > 0:
        path = out_path + out_prefix
        suff = True
    else:
        pos  = gff_path.rfind('/')
        if lsuf and suff and gzip:
            path = out_path + gff_path[pos:-8]
        elif suff and gzip:
            path = out_path + gff_path[pos:-7]
        elif suff and lsuf:
            path = out_path + gff_path[pos:-5]
        elif suff:
            path = out_path + gff_path[pos:-4]
        elif gzip:
            path = out_path + gff_path[pos:-3]
        else:
            path = out_path + gff_path[pos:]
        path += "." + datestamp

    if suff and gzip:
        path += ".gff.gz"
    elif gzip:
        path += ".gz"
    elif lsuf:
        path += ".gff3"
    elif suff:
        path += ".gff"

    if gzip:
        out_fh = gzip.open(path, "wb")
    else:
        out_fh = open(path, 'w')

    for chr in chromosomes:

        for k in range(0, len(chromosomes[chr].comps)):
            if chromosomes[chr].comps[k].type == "gap":
                continue

            for j in range(0, len(chromosomes[chr].comps[k].genes)):
                gcomps = ann[chromosomes[chr].comps[k].genes[j]]
                
                for g in gcomps:
                    attribs = []
                    for att in g.attribs:
                        attribs.append(att.key + "=\"" + att.value + "\"")

                    out_fh.write(g.chr      + "\t" +
                                 g.src      + "\t" +
                                 g.type     + "\t" +
                                 str(g.sbp) + "\t" +
                                 str(g.ebp) + "\t" +
                                 g.score    + "\t" +
                                 g.strand   + "\t" +
                                 g.phase    + "\t" +
                                 "; ".join(attribs) + "\n")
    out_fh.close()

def write_gtf(gtf_path, chromosomes, ann, datestamp):
    global out_path

    suff = False
    gzip = False
    if gtf_path[-7:] == ".gtf.gz":
        suff = True
        gzip = True
    elif gtf_path[-4:] == ".gtf":
        suff = True
    elif gtf_path[-3:] == ".gz":
        gzip = True

    if len(out_prefix) > 0:
        path = out_path + out_prefix
        suff = True
    else:
        pos  = gtf_path.rfind('/')
        if suff and gzip:
            path = out_path + gtf_path[pos:-7]
        elif suff:
            path = out_path + gtf_path[pos:-4]
        elif gzip:
            path = out_path + gtf_path[pos:-3]
        else:
            path = out_path + gtf_path[pos:]
        path += "." + datestamp

    if suff and gzip:
        path += ".gtf.gz"
    elif gzip:
        path += ".gz"
    elif suff:
        path += ".gtf"

    if gzip:
        out_fh = gzip.open(path, "wb")
    else:
        out_fh = open(path, 'w')

    for chr in chromosomes:

        for k in range(0, len(chromosomes[chr].comps)):
            if chromosomes[chr].comps[k].type == "gap":
                continue

            for j in range(0, len(chromosomes[chr].comps[k].genes)):
                gcomps = ann[chromosomes[chr].comps[k].genes[j]]

                for g in gcomps:
                    attribs = []
                    for att in g.attribs:
                        attribs.append(att.key + " \"" + att.value + "\"")

                    out_fh.write(g.chr      + "\t" +
                                 g.src      + "\t" +
                                 g.type     + "\t" +
                                 str(g.sbp) + "\t" +
                                 str(g.ebp) + "\t" +
                                 g.score    + "\t" +
                                 g.strand   + "\t" +
                                 g.phase    + "\t" +
                                 "; ".join(attribs) + "\n")
    out_fh.close()

def write_fasta(fasta_path, chromosomes, chromosomes_key, fasta_seqs, datestamp):
    global line_len, out_path, out_prefix

    suff = False
    lsuf = False
    gzip = False
    if fasta_path[-6:] == ".fa.gz":
        suff = True
        gzip = True
    elif fasta_path[-9:] == ".fasta.gz":
        suff = True
        gzip = True
        lsuf = True
    elif fasta_path[-6:] == ".fasta":
        suff = True
    elif fasta_path[-3:] == ".fa":
        suff = True
    elif fasta_path[-3:] == ".gz":
        gzip = True

    if len(out_prefix) > 0:
        path = out_path + out_prefix
        suff = True
    else:
        pos  = fasta_path.rfind('/')
        if lsuf and suff and gzip:
            path = out_path + fasta_path[pos:-9]
        elif suff and gzip:
            path = out_path + fasta_path[pos:-6]
        elif suff:
            path = out_path + fasta_path[pos:-3]
        elif gzip:
            path = out_path + fasta_path[pos:-3]
        else:
            path = out_path + fasta_path[pos:]
        path += "." + datestamp

    if suff and gzip:
        path += ".fa.gz"
    elif gzip:
        path += ".gz"
    elif suff:
        path += ".fa"

    if gzip:
        out_fh = gzip.open(path, "wb")
    else:
        out_fh = open(path, 'w')

    for chr in chromosomes:
        #
        # Build the sequence from the scaffolds.
        #
        tot_seq_len = 0
        for k in range(0, len(chromosomes[chr].comps)):
            if chromosomes[chr].comps[k].type != "gap":
                tot_seq_len += len(chromosomes[chr].comps[k].seq)
        if tot_seq_len == 0:
            continue

        seq = ""
        for k in range(0, len(chromosomes[chr].comps)):
            if chromosomes[chr].comps[k].type == "gap":
                seq += ('N' * chromosomes[chr].comps[k].len)
            else:
                seq += chromosomes[chr].comps[k].seq

        i = 0
        j = 0
        out_fh.write(">" + chr + "\n")
        num_lines = math.ceil(len(seq) / float(line_len))
        while j < num_lines:
            out_fh.write(seq[i:i+line_len] + "\n")
            i += line_len
            j += 1
        if j < num_lines:
            out_fh.write(seq[i:line_len] + "\n")

    #
    # We want to also include any scaffold-level sequences that weren't incorporated in the main assembly.
    #
    for id in fasta:
        if id not in chromosomes_key or chromosomes_key[id][0] != "groupUn":
            continue;

        seq = fasta[id]
        i   = 0
        j   = 0
        out_fh.write(">" + id + "\n")
        num_lines = math.ceil(len(seq) / float(line_len))
        while j < num_lines:
            out_fh.write(seq[i:i+line_len] + "\n")
            i += line_len
            j += 1
        if j < num_lines:
            out_fh.write(seq[i:line_len] + "\n")

    out_fh.close()

##------------------------------------------------------------------------------------------------##

parse_command_line()

#
# Parse the changes file.
#
changes = []
parse_changes_file(changes_path, changes)

scaf_agp     = {}
scaf_agp_key = {}
chr_agp      = {}
chr_agp_key  = {}

if len(scaf_agp_path) > 0:
    load_agp_file(scaf_agp_path, chr_agp, chr_agp_key)
if len(ctg_agp_path)> 0:
    load_agp_file(ctg_agp_path, scaf_agp, scaf_agp_key)
load_agp_file(chr_agp_path, chr_agp, chr_agp_key)

chromosomes     = {}
chromosomes_key = {}

#
# Put the chromosome, scaffold and contig data all in one place.
#
if len(ctg_agp_path) > 0:
    merge_contig_scaffold_agp(chromosomes, chromosomes_key, chr_agp, scaf_agp, scaf_agp_key)
else:
    merge_agp(chromosomes, chromosomes_key, chr_agp, scaf_agp, scaf_agp_key)

if unordered_scaffolds_concatenated == True and synthetic_chr not in chromosomes:
    print("Error: unable to identify synthetic chromosome ({}) holding unordered scaffolds.".format(syntentic_chr), file=sys.stderr)
    sys.exit()

#
# Load gene annotation information.
#
ann = {}

if len(gff_path) > 0:
    load_gff_file(gff_path, ann)
    merge_ann(chromosomes, ann)
if len(gtf_path) > 0:
    load_gtf_file(gtf_path, ann)
    merge_ann(chromosomes, ann)

fasta = {}

if len(fasta_path) > 0:
    load_fasta_file(fasta_path, fasta)
    merge_fasta(chromosomes, fasta)

#
# Process changes in the order read from the input changes file.
#
for i in range(0, len(changes)):
    op = changes[i]
    print("\nProcessing", op, file=sys.stderr)

    if len(chr_whitelist) > 0 and op.chr != chr_whitelist:
        continue

    if op.operation == "invert_scaffold":
        invert_scaffold(chromosomes, chromosomes_key, op, ann, fasta)
    elif op.operation == "invert":
        invert_chromosome(chromosomes, chromosomes_key, op, ann, fasta)
    elif op.operation == "insert_before":
        insert_scaffold_before(chromosomes, chromosomes_key, op, ann, fasta)
    elif op.operation == "insert_after":
        insert_scaffold_after(chromosomes, chromosomes_key, op, ann, fasta)
    elif op.operation == "fuse":
        fuse_chromosome(chromosomes, chromosomes_key, op, ann, fasta)
    elif op.operation == "split_after":
        split_chromosome_after(chromosomes, chromosomes_key, op, ann, fasta)
    elif op.operation == "rename":
        rename(chromosomes, chromosomes_key, op, ann, fasta)
    # print(chromosomes[chr])

#
# Write the results.
#
print("\nWriting spliced files...", end="", file=sys.stderr)
datestamp = datetime.now().strftime("%Y%m%d%H%M")

write_agp(chr_agp_path, chromosomes, datestamp)

if len(ann) > 0:
    if len(gff_path) > 0:
        write_gff(gff_path, chromosomes, ann, datestamp)
    else:
        write_gtf(gtf_path, chromosomes, ann, datestamp)

if len(fasta) > 0:
    write_fasta(fasta_path, chromosomes, chromosomes_key, fasta, datestamp)

print("done.", file=sys.stderr)
